from typing import Any, Dict, Type, Union

from django.test import TestCase, override_settings
from fake import FAKER, FILE_REGISTRY
from parametrize import parametrize

from ..base import DjangoBaseStorage
from ..filesystem import DjangoFileSystemStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2024-2025 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("TestStoragesTestCase",)


@override_settings(
    STORAGES={
        "default": {
            "BACKEND": (
                "django.core.files.storage.filesystem.FileSystemStorage"
            ),
            "OPTIONS": {},
        },
        "staticfiles": {
            "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
        },
    }
)
class TestStoragesTestCase(TestCase):
    """Test storages."""

    def tearDown(self) -> None:
        super().tearDown()
        FILE_REGISTRY.clean_up()  # Clean up files

    @parametrize(
        "storage_cls, kwargs, prefix, basename, extension",
        [
            # DjangoFileSystemStorage
            (
                DjangoFileSystemStorage,
                {
                    "root_path": "testing",
                    "rel_path": "tmp",
                },
                "zzz",
                None,
                "docx",
            ),
            (
                DjangoFileSystemStorage,
                {
                    "root_path": "testing",
                    "rel_path": "tmp",
                },
                None,
                "my_zzz_filename",
                "docx",
            ),
        ],
    )
    def test_storage(
        self: "TestStoragesTestCase",
        storage_cls: Type[DjangoBaseStorage],
        kwargs: Dict[str, Any],
        prefix: Union[str, None],
        basename: Union[str, None],
        extension: str,
    ) -> None:
        """Test storage."""
        storage = storage_cls(**kwargs)
        # Text file
        filename_text = storage.generate_filename(
            basename=basename, prefix=prefix, extension=extension
        )
        # Write to the text file
        text_result = storage.write_text(filename_text, "Lorem ipsum")
        # Check if file exists
        self.assertTrue(storage.exists(filename_text))
        # Assert correct return value
        self.assertIsInstance(text_result, int)
        # Clean up
        storage.unlink(filename_text)

        # Bytes
        filename_bytes = storage.generate_filename(
            basename=basename, prefix=prefix, extension=extension
        )
        # Write to bytes file
        bytes_result = storage.write_bytes(filename_bytes, b"Lorem ipsum")
        # Check if file exists
        self.assertTrue(storage.exists(filename_bytes))
        # Assert correct return value
        self.assertIsInstance(bytes_result, int)

        # Clean up
        storage.unlink(filename_bytes)

    @parametrize(
        "storage_cls, kwargs, prefix, extension",
        [
            # DjangoFileSystemStorage
            (
                DjangoFileSystemStorage,
                {
                    "root_path": "testing",
                    "rel_path": "tmp",
                },
                "zzz",
                "",
            ),
        ],
    )
    def test_storage_generate_filename_exceptions(
        self: "TestStoragesTestCase",
        storage_cls: Type[DjangoBaseStorage],
        kwargs: Dict[str, Any],
        prefix: str,
        extension: str,
    ) -> None:
        """Test storage `generate_filename` exceptions."""
        storage = storage_cls(**kwargs)

        with self.assertRaises(Exception):
            # Generate filename
            storage.generate_filename(prefix=prefix, extension=extension)

        with self.assertRaises(Exception):
            # Generate filename
            storage.generate_filename(basename=prefix, extension=extension)

    @parametrize(
        "storage_cls, kwargs, prefix, extension",
        [
            # DjangoFileSystemStorage
            (
                DjangoFileSystemStorage,
                {
                    "root_path": "root_tmp",
                    "rel_path": "rel_tmp",
                },
                "",
                "tmp",
            ),
        ],
    )
    def test_storage_abspath(
        self: "TestStoragesTestCase",
        storage_cls: Type[DjangoBaseStorage],
        kwargs: Dict[str, Any],
        prefix: str,
        extension: str,
    ) -> None:
        """Test `FileSystemStorage` `abspath`."""
        storage = storage_cls(**kwargs)
        filename = storage.generate_filename(
            prefix=prefix,
            extension=extension,
        )
        self.assertTrue(filename.startswith("root_tmp/rel_tmp/"))

    @parametrize(
        "storage_cls, kwargs, prefix, extension",
        [
            # DjangoFileSystemStorage
            (
                DjangoFileSystemStorage,
                {
                    "root_path": "root_tmp",
                    "rel_path": "rel_tmp",
                },
                "",
                "tmp",
            ),
        ],
    )
    def test_storage_unlink(
        self: "TestStoragesTestCase",
        storage_cls: Type[DjangoBaseStorage],
        kwargs: Dict[str, Any],
        prefix: str,
        extension: str,
    ) -> None:
        """Test `DjangoStorage` `unlink`."""
        storage = storage_cls(**kwargs)
        with self.subTest("Test unlink by Django"):
            filename_1 = storage.generate_filename(
                prefix=prefix,
                extension=extension,
            )
            storage.write_text(filename=filename_1, data=FAKER.text())
            self.assertTrue(storage.exists(filename_1))
            storage.unlink(filename_1)
            self.assertFalse(storage.exists(filename_1))

        with self.subTest("Test unlink by str"):
            filename_2 = storage.generate_filename(
                prefix=prefix,
                extension=extension,
            )
            storage.write_text(filename=filename_2, data=FAKER.text())
            self.assertTrue(storage.exists(filename_2))
            storage.unlink(str(filename_2))
            self.assertFalse(storage.exists(filename_2))



import os
from abc import abstractmethod
from pathlib import Path
from typing import Optional, Union

from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from fake import BaseStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2024-2025 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("DjangoBaseStorage",)

DEFAULT_ROOT_PATH = "tmp"
DEFAULT_REL_PATH = "tmp"


class DjangoBaseStorage(BaseStorage):
    """Django-based storage class using Django's default storage backend.

    Usage example:

    .. code-block:: python

        from fakepy.django_storage.filesystem import DjangoFileSystemStorage

        storage = DjangoFileSystemStorage()
        docx_file = storage.generate_filename(
            prefix="zzz_", extension="docx"
        )
        storage.write_bytes(docx_file, b"Sample bytes data")

    Initialization with params:

    .. code-block:: python

        from fakepy.django_storage.filesystem import DjangoFileSystemStorage

        storage = DjangoFileSystemStorage()
        docx_file = storage.generate_filename(
            prefix="example", extension="docx"
        )
    """

    def __init__(
        self: "DjangoBaseStorage",
        root_path: Optional[Union[str, Path]] = DEFAULT_ROOT_PATH,
        rel_path: Optional[str] = DEFAULT_REL_PATH,
        *args,
        **kwargs,
    ) -> None:
        """
        :param root_path: Root path of the storage directory.
        :param rel_path: Relative path of the storage directory.
        :param *args:
        :param **kwargs:
        """
        self.root_path = root_path or ""
        self.rel_path = rel_path or ""
        super().__init__(*args, **kwargs)
        self.storage = default_storage

    def generate_filename(
        self: "DjangoBaseStorage",
        extension: str,
        prefix: Optional[str] = None,
        basename: Optional[str] = None,
    ) -> str:
        """Generate filename."""
        dir_path = os.path.join(self.root_path, self.rel_path)

        if not extension:
            raise Exception("Extension shall be given!")

        if not basename:
            basename = self.generate_basename(prefix)

        filename = f"{basename}.{extension}"

        return os.path.join(dir_path, filename)

    def write_text(
        self: "DjangoBaseStorage",
        filename: str,
        data: str,
        encoding: Optional[str] = None,
    ) -> int:
        """Write text."""
        if filename.startswith("/"):
            filename = filename[1:]
        content = ContentFile(data.encode(encoding or "utf-8"))
        # saved_path = self.storage.save(filename, content)
        self.storage.save(filename, content)
        return len(data)

    def write_bytes(
        self: "DjangoBaseStorage",
        filename: str,
        data: bytes,
    ) -> int:
        """Write bytes."""
        content = ContentFile(data)
        if filename.startswith("/"):
            filename = filename[1:]
        # saved_path = self.storage.save(filename, content)
        self.storage.save(filename, content)
        return len(data)

    def exists(self: "DjangoBaseStorage", filename: str) -> bool:
        """Check if file exists."""
        return self.storage.exists(filename)

    @abstractmethod
    def relpath(self: "DjangoBaseStorage", filename: str) -> str:
        """Return relative path."""

    @abstractmethod
    def abspath(self: "DjangoBaseStorage", filename: str) -> str:
        """Return absolute path."""

    def unlink(self: "DjangoBaseStorage", filename: str) -> None:
        """Delete the file."""
        self.storage.delete(filename)



from google.cloud.exceptions import NotFound
from storages.backends.gcloud import GoogleCloudStorage
from storages.utils import clean_name

from .cloud import DjangoCloudStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2024-2025 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("DjangoGoogleCloudStorage",)


class DjangoGoogleCloudStorage(DjangoCloudStorage):
    """Django GoogleCloudStorage storage."""

    storage: GoogleCloudStorage

    def exists(self: "DjangoGoogleCloudStorage", filename: str) -> bool:
        """Check if file exists."""

        if not filename:  # root element aka the bucket
            try:
                self.storage.client.get_bucket(self.storage.bucket)
                return True
            except NotFound:
                return False

        name = self.storage._normalize_name(clean_name(filename))
        return bool(self.storage.bucket.get_blob(name))



from botocore.exceptions import ClientError
from storages.backends.s3 import S3Storage
from storages.utils import clean_name

from .cloud import DjangoCloudStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2024-2025 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("DjangoAWSS3Storage",)


class DjangoAWSS3Storage(DjangoCloudStorage):
    """Django AWS S3 storage."""

    storage: S3Storage

    def exists(self: "DjangoAWSS3Storage", filename: str) -> bool:
        """Check if file exists."""
        name = self.storage._normalize_name(clean_name(filename))

        try:
            self.storage.connection.meta.client.head_object(
                Bucket=self.storage.bucket_name, Key=name
            )
            return True
        except ClientError as err:
            if err.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
                return False

            # Some other error was encountered. Re-raise it.
            raise



from storages.backends.azure_storage import AzureStorage

from .cloud import DjangoCloudStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2024-2025 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("DjangoAzureCloudStorage",)


class DjangoAzureCloudStorage(DjangoCloudStorage):
    """Django AzureCloudStorage storage."""

    storage: AzureStorage

    def exists(self: "DjangoAzureCloudStorage", filename: str) -> bool:
        """Check if file exists."""
        if not filename:
            return True

        blob_client = self.storage.client.get_blob_client(
            self.storage._get_valid_path(filename)
        )
        return blob_client.exists()



from .base import DjangoBaseStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2024-2025 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("DjangoCloudStorage",)


class DjangoCloudStorage(DjangoBaseStorage):

    def relpath(self: "DjangoCloudStorage", filename: str) -> str:
        """Return relative path."""
        return filename

    def abspath(self: "DjangoCloudStorage", filename: str) -> str:
        """Return absolute path."""
        return self.storage.url(filename)



from .base import DjangoBaseStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2024-2025 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("DjangoFileSystemStorage",)


class DjangoFileSystemStorage(DjangoBaseStorage):
    """Django-based storage class using Django's default storage backend.

    Usage example:

    .. code-block:: python

        from fakepy.django_storage.filesystem import DjangoFileSystemStorage

        storage = DjangoFileSystemStorage()
        docx_file = storage.generate_filename(
            prefix="zzz_", extension="docx"
        )
        storage.write_bytes(docx_file, b"Sample bytes data")

    Initialization with params:

    .. code-block:: python

        from fakepy.django_storage.filesystem import DjangoFileSystemStorage

        storage = DjangoFileSystemStorage()
        docx_file = storage.generate_filename(
            prefix="example", extension="docx"
        )
    """

    def relpath(self: "DjangoFileSystemStorage", filename: str) -> str:
        """Return relative path."""
        return filename

    def abspath(self: "DjangoFileSystemStorage", filename: str) -> str:
        """Return absolute path."""
        return self.storage.joinpath(filename)
